;;;; This file is part of XMLTools
;;;; Copyright (C) 2004-2011, 2015 Sergey Poznyakoff
;;;;
;;;; XMLTools is free software; you can redistribute it and/or modify
;;;; it under the terms of the GNU General Public License as published by
;;;; the Free Software Foundation; either version 3 of the License, or
;;;; (at your option) any later version.
;;;;
;;;; XMLTools is distributed in the hope that it will be useful,
;;;; but WITHOUT ANY WARRANTY; without even the implied warranty of
;;;; MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
;;;; GNU General Public License for more details.
;;;;
;;;; You should have received a copy of the GNU General Public License
;;;; along with XMLTools.  If not, see <http://www.gnu.org/licenses/>.
;;;;

(define-module (xmltrans lingua)
  #:export (begin-processing parse-command-line end-processing))

(setlocale LC_ALL "")

(use-modules (xmltools xmltrans)
	     (xmltools gamlet)
	     (xmltools dict)
	     (ice-9 format)
             (ice-9 rdelim)
	     (ice-9 getopt-long)
	     (srfi srfi-13)
             (srfi srfi-14))

(cond
 ((= (string->number (major-version)) 1)
   (use-syntax (ice-9 syncase))
   (debug-enable 'debug)))
(debug-enable 'backtrace)

(define lingua-version "1.3")
(define lingua-version-string
  (string-append "lingua version "
		 lingua-version
		 " ($Id: lingua.scm 153 2015-01-11 15:23:51Z gray $)"))
(define lingua-copyright
  "Copyright (C) 2004-2011, 2015 Sergey Poznyakoff\n\
License GPLv3+: GNU GPL version 3 or later <http://gnu.org/licenses/gpl.html>\n\
This is free software: you are free to change and redistribute it.\n\
There is NO WARRANTY, to the extent permitted by law.\n\
\n\
Written by Sergey Poznyakoff.\n")

;;;; Command line options
(define base-href-option #f)
(define prefix-option #f)

(define config-file-name "lingua.conf")
(define config-must-exist #f)

(define output-dir-option #f)
(define tmp-dir-option #f)

(define defn-list '())

;;;; Global variables

;;; Keywords
(define keywords "")

;;; Index of the current item within the item group (e.g. ROW, ENUMERATE, etc)
(define item-index 0)
;;; Index of the current row within the row group (TABULAR)
(define row-index 0)
;;; Maximum number of items per row in the current table
(define max-items 0)
;;; Row index of the last separator row in the current table
(define separator-row 0)
;;; Keeps the index of the row ending the row span
(define rowspan-end 0)


;; Warning control
(define warning-list (list
		      #:author-data
		      #:page-title
		      #:page-header
		      #:unknown-modifier
		      #:prereq
; 		      #:soundref
;                      #:colalign-used
		      ))

(define (set-warning-control kw)
  (if (string-prefix? "no-" kw)
      (delq! (symbol->keyword (string->symbol (substring kw 3))) warning-list)
      (append! warning-list (list (symbol->keyword (string->symbol kw))))))


;;; Dictionary of file name translations
(define reference-translation-list '())

(define-macro (define-reference-translation file-name reference)
  `(set! reference-translation-list (cons
				     (cons ,file-name ,reference)
				     reference-translation-list)))
(define (file-name->reference name)
  (or (assoc-ref reference-translation-list name)
      name))


;;; Dictionary of cross-reference identifiers
;;; Each element is a cons:
;;;  (ID . #(FILENAME (cons LOCUS-FILE LOCUS-LINE)))
(define xref-id-list '())

;;; Retrieve a cross-reference object by its identifier
(define-macro (xref-retrieve id)
  `(assoc ,id xref-id-list))

;;; Create a cross-reference object
(define-syntax make-xref
  (syntax-rules ()
    ((make-xref id) (make-xref id
			   (current-file-name)
			   (xmltrans:get-input-location)))
    ((make-xref id file-name) (if (not file-name)
				(cons id #f)
				(make-xref id file-name
					 (xmltrans:get-input-location))))
    ((make-xref id file-name loc) (cons id (vector file-name loc)))))

(define-macro (xref-name id)
  `(car ,id))

;;; Return #t if xref ID is defined
(define-macro (xref-defined? id)
  `(cdr ,id))

;;; Return the target file name associated with the xref ID
(define-macro (xref-file-name id)
  `(vector-ref (cdr ,id) 0))

;;; Return the location of the definition of xref ID
(define-macro (xref-source-location id)
  `(vector-ref (cdr ,id) 1))

;;; Define an already registered cross-reference by associating with it
;;; the current file name and input location.
;;;
(define-macro (xref-define id)
  `(set-cdr! ,id (vector (current-file-name) (xmltrans:get-input-location))))

;;; Register a cross reference. Optional argument gives the target file name
;;; (or #f to create undefined (forward) cross-reference.
(define (xref-register id . rest)
  (cond
   ((assoc id xref-id-list) =>
    (lambda (x)
      (xmltrans:parse-error #f "Reference ID " (xref-name x) " already defined")
      (xmltrans:parse-error (xref-source-location x)
			    "This is the location of previos definition")))
   (else
    (set! xref-id-list
	  (cons
	   (cond
	    ((null? rest)
	     (make-xref id))
	    ((= (length rest) 1)
	     (make-xref id (car rest)))
	    (else
	     (throw 'misc-error 'menu-append
		    "Invalid number of optional arguments"
		    (list (xmltrans:input-file-name)
			  (xmltrans:input-line-number)) #f)))
	   xref-id-list)))))

;;; Resolve forward cross reference. This function is called from gamlet
;;; postprocess handler.
(define (resolve-forward-xref ref-name location)
  (let ((id (xref-retrieve ref-name)))
    (cond
     ((and id (xref-defined? id))
      (display (reference-name (file-name->reference (xref-file-name id))))
      (display "#")
      (display ref-name))
     (else
      (xmltrans:parse-error location "Undefined reference " ref-name)))))
      

;;;; Temporary gettext
(define gettext-trans '())

(define-public (lingua:gettext text)
  (or (assoc-ref gettext-trans text)
      text))

(xmltrans:end-tag
 "MSGSTR"
 (tag attrs text)
 (let ((id (xmltrans:attr #f "ID")))
   (if id
       (set! gettext-trans (cons
			    (cons id text)
			    gettext-trans))
       (xmltrans:parse-error #f "MSGSTR tag missing ID")))
 #f)


;;;; Menu support
(define (menu-append tag object label)
  (let ((menu (xmltrans:attr tag "MENU")))
    (cond
     ((not menu)
      (throw 'misc-error 'menu-append
	     "~A:~A: Parent element does not have MENU element"
	     (list (xmltrans:input-file-name)
		   (xmltrans:input-line-number)) #f)))
    (xmltrans:set-attr
     tag "MENU"
     (append menu
	     (list (cons object label))))))
		       

(define (find-file-in-menu filename . rest)
  (let ((menu (if (null? rest) (lingua:attr "LINGUA" "MENU") (car rest))))
    (call-with-current-continuation
     (lambda (return)
       (for-each
	(lambda (elem)
	  (let ((object (car elem)))
	    (cond
	     ((string? object)
	      (if (string=? object filename)
		  (return (cdr elem))))
	     ((list? object)
	      (let ((x (find-file-in-menu filename object)))
		(if x
		    (return x)))))))
	menu)
       #f))))

(define (print-menu-map menu)
  (cond
   ((not (null? menu))
    (display "<ul>\n")
    (for-each
     (lambda (elem)
       (let ((object (car elem))
	     (label (cdr elem)))
	 (cond
	  ((string? object)
	   (format #t "<li><a href=\"~A\">~A</a></li>"
		   (reference-name object)
		   label))
	  ((list? object)
	   (cond
	    ((string-null? label)
	     (format #t "<li><a href=\"~A\">~A</a>\n"
		     (reference-name (caar object)) (cdar object))
	     (print-menu-map (cdr object)))
	    (else
	     (display "<li>")
	     (display label)
	     (print-menu-map object)))
	   (display "</li>")))))
     menu)
    (display "</ul>"))))

(define (current-menu-item? item)
  (and
   (string? (car item))
   (string=? (car item) (file-name->reference
			 (gamlet:current-output-file-name)))))

(define (put-menu-item item level)
  (letrec ((put-string-item (lambda (item)
			      (cond
			       ((current-menu-item? item)
				(format #t "<span class=\"active\">~A</span>"
					(cdr item)))
			       (else
				(format #t "<a href=\"~A\">~A</a>"
					(reference-name (car item))
					(cdr item)))))))
    (cond
     ((string? item)
      (display "<li>")
      (display item)
      (display "</li>"))
     ((procedure? (car item))
      ((car item)))
     ((list? (car item))
      (if (and (string? (cdr item)) (not (string-null? (cdr item))))
	  (format #t
		  "<li class=\"group\"><div class=\"group\">~A</div>\n"
		  (cdr item))
	  (display "<li>"))
      (cond
       ((= 0 level)
	(out-menu (car item) (1+ level)))
       (else
	(put-string-item (caar item))
	(if (has-current-menu-item? (car item))
	    (out-menu (cdr (car item)) (1+ level)))))
      (display "</li>"))
     ((string? (car item))
      (display "<li>")
      (put-string-item item)
      (display "</li>\n"))
     (else
      (throw 'misc-error 'put-menu-item
	     "Unrecognized menu item ~S" (list item) #f)))))
  
(define (has-current-menu-item? menu)
  (call-with-current-continuation
   (lambda (exit)
     (for-each
      (lambda (item)
	(cond
	 ((list? (car item))
	  (if (has-current-menu-item? (car item))
	      (exit #t)))
	 (else
	  (if (current-menu-item? item)
	      (exit #t)))))
      menu)
     #f)))
     

(define (out-menu menu level)
;  (display "MENU " (current-error-port))
;  (write menu (current-error-port))
;  (display "\n" (current-error-port))
  (display "<ul class=\"menu\">")
  (for-each
   (lambda (item)
     (put-menu-item item level))
   menu)
  (display "</ul>"))
	      
		


;;;; Utility functions

(define-macro (defn-put func-name arg-name)
  `(lambda ( . rest )
		(cond
		 ((not (null? rest))
		  (set! ,arg-name (cons (car rest) ,arg-name))
		  (apply ,func-name (cdr rest))))))

(define (left-pad-string str n chr)
  (let ((len (string-length str)))
    (cond
     ((< len n)
      (string-append (make-string (- n len) chr) str))
     (else
      str))))

(define (create-file-name)
  (let ((lang (let ((l (xmltrans:attr "LINGUA" "LANG")))
		(if l
		    (string-append l "/")
		    "")))
	(suffix ".html")
	(prefix (or (get-attribute-value #f "PREFIX") "prefix"))
	(mode (string->symbol (or (get-attribute-value #f "MODE")
				  "numbered"))))
    (case mode
      ((numbered)
       (cond
	((get-attribute-value #f "PAGENO" 1+) =>
	 (lambda (n)
	   (if (= 0 n)
	       (string-append lang prefix suffix)
	       (string-append lang prefix "_"
			      (left-pad-string (number->string n) 2 #\0)
			      suffix))))
	(else
	 (string-append lang prefix suffix))))
      ((unnumbered)
       (string-append lang prefix suffix))
      (else
       (xmltrans:parse-error #f "Unknown mode " mode)))))

;;;; Look up for attribute NAME in the current tag stack starting
;;;; from TAG or from the stack top if TAG is #f. Return attribute
;;;; value if found or #f otherwise
(define (get-attribute-value tag name . rest)
  (let ((proc (and (not (null? rest)) (car rest))))
    (call-with-current-continuation
     (lambda (exit)
       (letrec ((getval
		 (lambda (name attrs)
		   (let ((item (assoc name attrs)))
		     (if item
			 (let ((val (cdr item)))
			   (if proc
			       (set-cdr! item (proc val)))
			   (exit val)))))))
	 (xmltrans:for-each-tag
	  (lambda (tag-name tag-attrs)
	    (if tag
		(cond
		 ((string=? tag-name tag)
		  (set! tag #f)
		  (getval name tag-attrs)))
		;; else:
		(getval name tag-attrs)))))
       #f))))

(define (reference-name name)
; FIXME: Reconsider this logic. In particular, the (tgt) branch below
; which is true even if tgt="". That's OK, but the logic looks strange. 
  (let ((tgt (or (and (xmltrans:tag-exists? "LINGUA")
		      (xmltrans:attr "LINGUA" "TARGETDIR"))
		 (and (lingua:tag-exists? "LINGUA")
		      (lingua:attr "LINGUA" "TARGETDIR"))
		 (and (lingua:tag-exists? "DIRECTORY")
		      (lingua:attr "DIRECTORY" "TARGET")))))
    (cond
     ((char=? (string-ref name 0) #\/)
      (if (and tgt (string-any char-set:graphic tgt))
	  (substring name 1)
	  name))
     (tgt
      (string-append tgt "/" name))
     (prefix-option
      (string-append prefix-option "/" name))
     (else
      name))))

(define (image-reference)
  (or (and (lingua:tag-exists? "DIRECTORY")
	   (lingua:attr "DIRECTORY" "GRAPHICS"))
      (reference-name "graphics")))
	 
(define (current-file-name)
  (xmltrans:attr "PAGE" "FILE-NAME"))

;;;; Footnote support
(define (add-footnote obj text id)
  (cond
   ((and id (assoc id (or (xmltrans:attr obj "FOOTNOTE-IDS") '()))) =>
    (lambda (n)
      (cdr n)))
   (else
    (let ((fl (or (xmltrans:attr obj "FOOTNOTES") '())))
      (xmltrans:set-attr obj "FOOTNOTES" (append fl (list text)))
      (let ((fn (1+ (length fl))))
	(if id
	    (xmltrans:set-attr obj
			       "FOOTNOTE-IDS"
			       (cons
				(cons id fn)
				(or (xmltrans:attr obj "FOOTNOTE-IDS") '()))))
	fn)))))

(define (footnote-name n)
  (string-append "FNNAME:" (number->string n)))

(define (footnote-href n)
  (string-append "FNHREF:" (number->string n)))


(define-macro (map-footnotes fl proc)
  `(cond
    (,fl
     (,proc "<div class=\"footnotes\">")
     (do ((cur ,fl (cdr cur))
	  (n 1 (1+ n)))
	 ((null? cur))
       (,proc (format #f "<p><a name=\"~A\" href=\"~A#~A\"><sup>~A)</sup></a> "
		      (footnote-href n)
		      (reference-name (file-name->reference (current-file-name)))
		      (footnote-name n)
		      n))
       (,proc (car cur))
       (,proc "</p>\n"))
     (,proc "</div>\n"))))

(define (print-footnotes fl)
  (map-footnotes fl display))
	   
(define-public (lingua:get-cgi-bin)
  (or
   (and (lingua:tag-exists? "DIRECTORY")
	(lingua:attr "DIRECTORY" "CGIBIN"))
   "/cgi-bin"))

(define-public (lingua:get-keyword-value args keyword default)
  (let ((kv (memq keyword args)))
    (if (and kv (>= (length kv) 2))
	(cadr kv)
	default)))

(define-public (lingua:LANG args)
  (let ((lang (or (lingua:get-keyword-value args #:lang #f) 
		  (xmltrans:attr "LINGUA" "LANG")))
	(ter (lingua:get-keyword-value args #:ter '())))
    (if (null? ter)
	lang
	(string-append
	 lang "_" (or ter (xmltrans:attr "LINGUA" "TER"))))))


(define (get-email)
  (cond
   ((lingua:attr "AUTHOR" "EMAIL") =>
    (lambda (x)
      x))
   (else
    "gray@gnu.org")))

(define (get-base-href)
  (cond
   ((xmltrans:attr "LINGUA" "BASE") =>
    (lambda (x)
      x))
   (else
    base-href-option)))

(define (open-page title)
  (let ((lang (xmltrans:attr "LINGUA" "LANG")))
    (format #t "\
<!DOCTYPE html PUBLIC \"-//W3C//DTD XHTML 1.0 Strict//EN\" 
          \"http://www.w3.org/TR/xhtml1/DTD/xhtml1-strict.dtd\">
<html xmlns=\"http://www.w3.org/1999/xhtml\" xml:lang=\"~A\" lang=\"~A\">
<head>
 <title>~A</title>
 <link rev=\"made\" href=\"mailto:~A\" />
 <link rel=\"stylesheet\" type=\"text/css\" href=\"~A\" />\n"
	    lang
	    lang
	    title
	    (get-email)
	    (reference-name "style.css"))
    (let ((linkgen (xmltrans:attr "LINGUA" "LINK")))
      (cond
       ((not linkgen))
       ((procedure? linkgen)
	(linkgen))
       ((string? linkgen)
	(for-each
	 (lambda (expr)
	   (let ((x (string-split expr #\=)))
	     (cond
	      ((not (= (length x) 2))
	       (xmltrans:parse-error #f "LINK syntax: too many equal signs" expr))
	      ((string=? (car x) "rss")
	       (format #t
		       "<link rel=\"alternate\" title=\"~A - RSS\" href=\"~A\" type=\"application/rss+xml\" />\n"
		       title
		       (cadr x)))
	      ((string=? (car x) "icon")
	       (format #t "<link rel=\"shortcut icon\" href=\"~A\" />\n"
		       (cadr x)))
	      (else
	       (xmltrans:parse-error #f "LINK syntax: unknown keyword: "
				     (car x))))))
	 (string-split linkgen #\;)))
       (else
	(xmltrans:parse-error #f "LINK syntax: " linkgen))))

    (format #t "\
 <meta http-equiv=\"Content-Type\" content=\"text/html; charset=UTF-8\" />
 <meta name=\"Generator\" content=\"~A; ~A\" />
 <meta name=\"Keywords\" lang=\"~A\" content=\"~A\" />
 <meta name=\"robots\" content=\"~A\" />
"
	    xmltrans:generator
	    lingua-version-string
	    lang
	    keywords
	    (or (xmltrans:attr #f "ROBOTS")
		"index,follow")))

  (let ((base (get-base-href)))
    (if (and base (not (string-null? base)))
	(format #t " <base href=\"~A\" />~%" base)))

  (display "\
</head>
<body>
<div class=\"main\">
")

  (let ((warning (or (xmltrans:attr #f "WARNING")
		     (and (xmltrans:tag-exists? "SECTION")
			  (xmltrans:attr "SECTION" "WARNING"))
		     (and (xmltrans:tag-exists? "CHAPTER")
			  (xmltrans:attr "CHAPTER" "WARNING"))
		     (and (lingua:tag-exists? "WARNING")
			  (lingua:tag-text "WARNING")))))
    (if (and warning (string-any char-set:graphic warning))
	(format #t "
<div class=\"warning\">
~A
</div>"
		warning)))

  (format #t "
 <div class=\"contentpanel\">
  <div class=\"content-inner\">	  
   <h1>~A</h1>~%"

	  (or (xmltrans:attr #f "HEADER") title)))
  
(define (close-page)
  (print-footnotes (xmltrans:attr "PAGE" "FOOTNOTES"))
  (if (not (get-attribute-value #f "HIDDEN"))
      (display "{BOTTOM-NAVBAR}"))
  (format #t "
  </div> <!-- content-inner -->
 </div> <!-- contentpanel -->

 <div class=\"navbar\">
  <img src=\"~A\" alt=\"~A\" />

   {TOC}

   <ul class=\"menu\">
    <li class=\"group add-info\"><div class=\"group\">~A</div>
    <ul class=\"menu\"> 
     <li><a href=\"~A\">~A</a></li>
     <li><a href=\"http://validator.w3.org/check/referer\" class=\"imgref\">
      <img src=\"~A/valid-xhtml10.png\"
          alt=\"[ Valid XHTML 1.0! ]\" height=\"31\" width=\"88\" /></a></li>

     <li><a href=\"http://jigsaw.w3.org/css-validator/check/referer\" class=\"imgref\">
      <img src=\"~A/valid-css.png\" 
         alt=\"[ Valid CSS! ]\" height=\"31\" width=\"88\" /></a></li>

     <li><a href=\"http://www.gnu.org\" class=\"imgref\">
      <img src=\"~A/button-powered_by_gnu.png\"
         alt=\"[ Powered by GNU ]\" height=\"47\" width=\"102\" /></a></li>
    </ul></li>
   </ul>

   ~A
	  
 </div> <!-- navbar -->
</div> <!-- main -->
 "
	  (reference-name (lingua:attr "LOGO" "SRC"))
	  (lingua:attr "LOGO" "ALT")
	  (lingua:gettext "Information")
	  (lingua:attr "AUTHOR" "HOMEPAGE")
	  (lingua:gettext "Home Page")
	  (image-reference)
	  (image-reference)
	  (image-reference)
	  (cond
	   ((xmltrans:attr "PAGE" "REF") =>
	    (lambda (x)
	      (string-append "{INFO \"" x "\"}")))
	   (else
	    "{INFO}")))

  (display "<div class=\"bottom\">\n")
  (display "<div class=\"copyright\">\n")
  (cond
   ((xmltrans:attr "PAGE" "BOTTOM") =>
    (lambda (x)
      (newline)
      (display x)
      (newline))))

  (cond
   ((or (xmltrans:attr "PAGE" "COPYRIGHT")
	(and (xmltrans:tag-exists? "SECTION")
	     (xmltrans:attr "SECTION" "COPYRIGHT"))
	(and (xmltrans:tag-exists? "CHAPTER")
	     (xmltrans:attr "CHAPTER" "COPYRIGHT"))) =>
    (lambda (x)
      (display x)))
   ((and (lingua:tag-exists? "COPYRIGHT")
	 (lingua:attr "COPYRIGHT" "TEXT")) =>
	 (lambda (x)
	   (display x)))
   ((lingua:attr "AUTHOR" "CONTACT") =>
    (lambda (x)
      (format #t "<p>~A <a href=\"mailto:~A\">~A</a>;</p>"
	      (lingua:gettext "For information about this site contact")
	      (get-email)
	      x)))
   (else
    (if (memq #:author-data warning-list)
	(xmltrans:parse-warning #f
				"Neither AUTHOR.CONTACT nor COPYRIGHT are defined"))))
      
  (display "</div> <!-- copyright -->\n
</div> <!-- bottom -->\n
</body>
</html>"))


(define (set-anchor attrs . rest)
  (cond
   ((xmltrans:attr attrs "ID") =>
    (lambda (xref-name)
      (let ((id (xref-retrieve xref-name)))
	(if id
	    (xref-define id)
	    (xref-register xref-name)))
      (format #f "<a name=\"~A\">~A</a>"
	      xref-name
	      (if (not (null? rest)) (car rest) ""))))
   (else
    #f)))



;;;; Handlers

;;;; Default handlers

(define tag-list '())

(define (lingua:default-start tag attr)
  (xmltrans:set-attr #f "__START__" 1)
  #f)

(xmltrans:set-default-start-handler lingua:default-start)

(define (lingua:default-end tag attr text)
  (if (xmltrans:attr attr "__START__")
      (xmltrans:parse-error #f "Unhandled element " tag))
  (set! tag-list
	(cons
	 (xmltrans:make-tag tag attr text)
	 tag-list))
  #f)

(xmltrans:set-default-end-handler lingua:default-end)

(define (lingua:handled-element-end tag attr text)
  (set! tag-list
	(cons
	 (xmltrans:make-tag tag attr text)
	 tag-list))
  #f)

(define (lingua:merge-element-end tag attr text)
  (cond
   ((assoc tag tag-list) =>
    (lambda (x)
      (xmltrans:tag-set! #:attrs x
			 (append attr (xmltrans:tag-ref #:attrs x)))
      (xmltrans:tag-set! #:text x (string-append
				   (xmltrans:tag-ref #:text x) text))))
   (else
    (set! tag-list
	  (cons
	   (xmltrans:make-tag tag attr text)
	   tag-list))))
  #f)

(define-syntax lingua:tag-ignore
  (syntax-rules ()
    ((lingua:tag-ignore name)
     (xmltrans:set-end-handler (symbol->string 'name)
			       lingua:handled-element-end))
    ((lingua:tag-ignore name 'merge)
     (xmltrans:set-end-handler (symbol->string 'name)
			       lingua:merge-element-end))))

;; FIXME: duplicated from xmltrans
;;;; Returns #t if TAG exists on stack
(define-public (lingua:tag-exists? tag)
  (assoc tag tag-list))

(define-public (lingua:tag-attributes tag)
  (let ((item (if tag (assoc tag tag-list) (car tag-list))))
    (if (not item)
	(throw 'misc-error 'lingua:tag-attributes
	       "No such tag on stack ~S" (list tag) #f))
    (xmltrans:tag-ref #:attrs item)))

(define-public (lingua:tag-text tag)
  (let ((item (if tag (assoc tag tag-list) (car tag-list))))
    (if (not item)
	(throw 'misc-error 'lingua:tag-attributes
	       "No such tag on stack ~S" (list tag) #f))
    (xmltrans:tag-ref #:text item)))

(define-public (lingua:attr tag name)
  (let ((attr (assoc name (lingua:tag-attributes tag))))
    (and attr (cdr attr))))

(define-public (lingua:set-attr tag name value)
  (let ((item (xmltrans:extract-item tag-list tag)))
    (if (not item)
	(throw 'misc-error 'lingua:set-attr
	       "No such tag on stack ~S" (list tag) #f))
    (let ((attr (assoc name (xmltrans:tag-ref #:attrs item))))
      (if attr
	  (set-cdr! attr value)
	  (xmltrans:tag-set! #:attrs
			     item
			     (cons
			      (cons name value)
			      (xmltrans:tag-ref #:attrs item)))))))


(define (replicate-menu from to)
  (cond
   ((xmltrans:attr from "MENU") =>
    (lambda (m)
      (xmltrans:set-attr to "MENU" m)))))


;;;; Element handlers

;;; Configuration file elements
(xmltrans:end-tag
 "DIRECTORY"
 (tag attr text)
 (for-each
  (lambda (a)
    (if (not (lingua:tag-exists? "DIRECTORY"))
	(set! tag-list (cons
			(xmltrans:make-tag "DIRECTORY")
			tag-list)))
    (lingua:set-attr "DIRECTORY" (car a) (cdr a)))
  attr)
 #f)

(xmltrans:end-tag
 "BASE"
 (tag attr text)
 (if (not base-href-option)
     (set! base-href-option (xmltrans:attr attr "HREF")))
 #f)

(xmltrans:end-tag
 "DICT"
 (tag attr text)
 (for-each
  (lambda (a)
    (if (not (lingua:tag-exists? "DICT"))
	(set! tag-list (cons
			(xmltrans:make-tag "DICT")
			tag-list)))
    (lingua:set-attr "DICT" (car a) (cdr a)))
  attr)
 #f)

;;; Ignored tags
(lingua:tag-ignore LOGO)
(lingua:tag-ignore CONFIG)
(lingua:tag-ignore AUTHOR 'merge)
(lingua:tag-ignore LOCALIZATION)
(lingua:tag-ignore SOUNDREF)
(lingua:tag-ignore SEARCH)
(lingua:tag-ignore WARNING)
 
;;; Simple tags
(xmltrans:replace-tag DFN dfn)
(xmltrans:replace-tag CMD code)
(xmltrans:replace-tag PROG b)
(xmltrans:replace-tag QUOTE i)
(xmltrans:replace-tag ITEMIZE dl)
(xmltrans:replace-tag PREF b)
(xmltrans:replace-tag SUFF b)
(xmltrans:replace-tag EMPH b)
(xmltrans:replace-tag SUBSCRIPT sub)
(xmltrans:replace-tag SUPERSCRIPT sup)
(xmltrans:replace-tag IT i)

(xmltrans:end-tag
 "NEWLINE"
 (tag attr text)
 (if (string-null? text)
     "\n"
     (list "\n" text "\n")))

;;;; INCLUDE
(xmltrans:end-tag
 "INCLUDE"
 (tag attr text)
 (let ((fname (xmltrans:attr attr "FILE")))
   (cond
    ((not fname)
     (xmltrans:parse-error #f "File name not specified"))
    ((string? fname)
     (replicate-menu #t #f)
     (xmltrans:parse-file fname)
     (replicate-menu #f #t))
    (else
     (xmltrans:parse-error #f "FILE must be a valid string"))))
 #f)

;;;; COPYRIGHT
(xmltrans:end-tag
 "COPYRIGHT"
 (tag attr text)
 (cond
  ((xmltrans:parent? "PAGE")
   (xmltrans:set-attr "PAGE" "COPYRIGHT" text))
  ((xmltrans:parent? "SECTION")
   (xmltrans:set-attr "SECTION" "COPYRIGHT" text))
  ((xmltrans:parent? "CHAPTER")
   (xmltrans:set-attr "CHAPTER" "COPYRIGHT" text))
  (else
   (lingua:handled-element-end tag (cons (cons "TEXT" text) attr) text)))
 #f)

;;;; GUILE
(xmltrans:end-tag
 "GUILE"
 (tag attr text)
 (cond
  ((xmltrans:attr attr "SOURCE") =>
   (lambda (f)
     (load f))))
 (if (not (string-null? text))
     (eval-string text (resolve-module '(xmltrans lingua)))))

;;;;; Conditionals
(define-public (lingua:getval name)
  (assoc-ref defn-list name))

(define-public (lingua:define name val)
  (set! defn-list (cons (cons name val) defn-list)))

(define-public (lingua:undefine name)
  (set! defn-list
	(filter (lambda (elt)
		  (not (string=? (car elt) name))) defn-list)))

(xmltrans:end-tag
 "DEFINE"
 (tag attr text)
 (cond
  ((xmltrans:attr attr "NAME") =>
   (lambda (name)
     (lingua:define name text)))
  (else
   (xmltrans:parse-error #f "Missing NAME attribute")))
 #f)

(xmltrans:end-tag
 "UNDEFINE"
 (tag attr text)
 (cond
  ((xmltrans:attr attr "NAME") =>
   (lambda (name)
     (lingua:undefine name)))
  (else
   (xmltrans:parse-error #f "Missing NAME attribute")))
 #f)
 
(xmltrans:end-tag
 "IF"
 (tag attr text)
 (if (cond
      ((xmltrans:attr attr "DEF") =>
       (lambda (v)
	 (cond
	  ((lingua:getval v) =>
	   (lambda (def)
	     (cond
	      ((xmltrans:attr attr "EQ") =>
	       (lambda (val)
		 (and (string? def)
		      (string=? val def))))
	      (else
	       #t))))
	  (else
	   #f))))
      ((xmltrans:attr attr "NDEF") =>
       (lambda (v)
	 (not (lingua:getval v))))
      (else
       #f))
     text
     #f))

(xmltrans:end-tag
 "VALUE"
 (tag attr text)
 (cond
  ((xmltrans:attr attr "NAME") => lingua:getval)))

;;;; LINGUA handlers
(define-public (lingua:expand-additional-inputs attr str)
  (map
   (lambda (inp)
     (let ((p (string-split inp #\=)))
       (string-append
	"<input type=\"hidden\" name=\"" (car p)
	"\" value=\""
	(apply
	 string-append
	 (map
	  (lambda (s)
	    (cond
	     ((char=? (string-ref s 0) #\@)
	      (xmltrans:attr attr (substring s 1)))
	     ((char=? (string-ref s 0) #\_)
	      (lingua:gettext (substring s 1)))
	     (else s)))
	  (string-tokenize (cadr p)
			   (char-set-complement (->char-set "{}")))))
	"\">")))
   (string-split str #\,)))

(define (compare-version va vb)
  (call-with-current-continuation
   (lambda (return)
     (letrec ((version->list
	       (lambda (str)
		 (catch #t
			(lambda ()
			  (map
			   string->number
			   (string-split str #\.)))
			(lambda (key . args)
			  (return #f)))))
	      (augment-version-list
	       (lambda (arg minlen)
		 (let ((len (length arg)))
		   (if (< len minlen)
		       (append arg (make-list (- minlen len) 0))
		       arg)))))

       (let* ((valst (version->list va))
	      (valen (length valst))
	      (vblst (version->list vb))
	      (vblen (length vblst))
	      (maxlen (max valen vblen)))
	 (for-each
	  (lambda (a b)
	    (cond
	     ((< a b)
	      (return -1))
	     ((> a b)
	      (return 1))))
	  (augment-version-list valst maxlen)
	  (augment-version-list vblst maxlen))

	 (return 0))))))

(define (prereq-commit arg)
  (let ((required-id
	 (catch #t
		(lambda ()
		  (string->number arg))
		(lambda (key . args)
		  (xmltrans:parse-error #f "Invalid commit number in PREREQ: " arg)
		  (xmltrans:parse-error #f "Cannot compare commit numbers, aborting")
		  (exit 1))))
	(commit-id (string->number
		    (list-ref
		     (string-split
		      (substring lingua-version-string
				 (string-contains lingua-version-string
						  "$Id: lingua.scm "))
		      #\space)
		     2))))
    (cond
     ((< commit-id required-id)
      (xmltrans:parse-error #f "Lingua commit number too low (" commit-id
			    ", required at least " required-id ")")
      (exit 1)))))
    
(define (lingua-prereq val)
  (for-each
   (lambda (arg)
     (cond
      ((string-null? arg)) ; Skip multiple spaces
      ((string-prefix? "commit=" arg)
       (prereq-commit (substring arg 7)))
      ((compare-version arg lingua-version) =>
       (lambda (res)
	 (cond
	  ((not res)
	   (xmltrans:parse-error #f "Invalid version number in PREREQ: " arg)
	   (xmltrans:parse-error #f "Cannot compare version numbers, aborting")
	   (exit 1))
	  ((> res 0)
	   (xmltrans:parse-error #f "Lingua version too old (" lingua-version
	    ", required at least " arg ")")
	   (exit 1)))))))
   (string-split val #\space)))

(xmltrans:start-tag
 "LINGUA"
 (tag attr)
 (cond
  ((xmltrans:attr "LINGUA" "PREREQ") => lingua-prereq)
  ((memq #:prereq warning-list)
   (xmltrans:parse-warning #f "Suggested the use of LINGUA PREREQ attribute")))
  ;(set-locale (xmltrans:attr "LINGUA" "LANG"))
 (letrec ((dictref (lambda (. args)
		     (string-append
		      (lingua:get-cgi-bin)
		      "/"
		      (or (and (lingua:tag-exists? "DICT")
			       (lingua:attr "DICT" "PROG"))
			  "dict.cgi")
		      "?lang="
		      (lingua:LANG args))))
	  (search (lambda (. args)
		    (put-menu-item
		     (or
		      (and (lingua:tag-exists? "SEARCH")
			   (not (string-null? (lingua:tag-text "SEARCH")))
			   (let ((text (lingua:tag-text "SEARCH"))
				 (args (lingua:attr "SEARCH" "ARGS")))
			     (cond
			      (args
			       (apply format #f text 
				      (eval-string args (resolve-module '(xmltrans lingua)))))
			      (else
			       text))))
				  
		      (let ((htsearch "/cgi-bin/htsearch")
			    (input-width 30)
			    (add-inputs '()))
			(if (lingua:tag-exists? "SEARCH")
			    (begin
			      (set! htsearch
				    (or (lingua:attr "SEARCH" "HTSEARCH")
					htsearch))
			      (set! input-width
				    (or (lingua:attr "SEARCH" "INPUT-WIDTH")
					input-width))
			      (set! add-inputs
				    (if (lingua:attr "SEARCH" "ADD-INPUTS")
					(lingua:expand-additional-inputs
					 attr
					 (lingua:attr "SEARCH" "ADD-INPUTS"))
					add-inputs))))
			  
			(format #f 
			    "<div class=\"htdig\">
                             <form method=\"post\" action=\"~A\">
			     ~A
			     <input type=\"text\" size=\"~A\" name=\"words\" value=\"\" />
			     ~{~A~} 
                             </form>
	                     </div>"
			    htsearch
			    (lingua:gettext "Search")
			    input-width
			    add-inputs)))
		     0)))
	  
	  (navigation
	   (lambda ()
	     (let ((index (gamlet:output-file-index)))
	     (if (> index 0)
		 (put-menu-item (cons
				 (file-name->reference
				  (gamlet:output-file-name (- index 1)))
				 (lingua:gettext "Previous Page"))
				0))
	     (if (< index (- (gamlet:output-file-count) 1))
		 (put-menu-item (cons
				 (file-name->reference
				  (gamlet:output-file-name (+ index 1)))
				 (lingua:gettext "Next Page"))
				0))))))
			  
   (xmltrans:set-attr #f "DICT" dictref)
   (xmltrans:set-attr #f "SEARCH" search)
   (xmltrans:set-attr #f "NAVIGATION" navigation))
 
 (xmltrans:set-attr #f "MENU" '())
 "")

(xmltrans:end-tag
 "LINGUA"
 (tag attr text)
 (lingua:default-end tag attr text)
 #f)

;;; 

;;; PAGE

(define (page-title)
  (cond
   ((or (xmltrans:attr "PAGE" "TITLE")
	(xmltrans:attr "PAGE" "HEADER")) =>
	(lambda (x)
	  x))
   (else
    (if (memq #:page-title warning-list)
	(xmltrans:parse-warning #f "Page title not defined"))
    (string-append "UNDEFINED at "
		   (xmltrans:input-file-name)
		   ":"
		   (number->string
		    (xmltrans:input-line-number))))))

(define (page-menu tag)
  (menu-append
   tag
   (cond
    ((xmltrans:attr "PAGE" "REF") =>
     (lambda (x)
       (let ((fn (xmltrans:attr "LINGUA" x)))
	 (cond
	  ((not fn)
	   (xmltrans:parse-error #f "Function " x " not defined")
	   (current-file-name))
	  ((not (procedure? fn))
	   (xmltrans:parse-error #f x " is not a function")
	   (current-file-name))
	  (else
	   (let ((ref (fn)))
	     (define-reference-translation (current-file-name) ref)
	     ref))))))
    (else
     (current-file-name)))

   ;; label:
   (or (xmltrans:attr "PAGE" "TITLE")
       (xmltrans:attr "PAGE" "HEADER")
       
       (begin
	 (if (memq #:page-header warning-list)
	     (xmltrans:parse-warning #f "Page header not defined"))
	 (string-append "UNDEFINED at "
			(xmltrans:input-file-name)
			":"
			(number->string
			 (xmltrans:input-line-number)))))))

(xmltrans:start-tag
 "PAGE"
 (tag attr)
 (xmltrans:set-attr
  #f
  "FILE-NAME"
  (create-file-name))
 (if (xmltrans:attr attr "HEADER")
     (page-menu #t)
     (let ((loc (xmltrans:get-input-location)))
       (letrec ((proc (lambda ()
			(if (memq #:page-header warning-list)
			    (xmltrans:parse-warning
			     loc
			     "PAGE element does not have HEADER attribute."))
			(xmltrans:set-attr "PAGE" "HEADER"
					   (xmltrans:attr "PAGE" "TITLE"))
			(xmltrans:set-attr "PAGE" "__HEADER_INSERTED__" 1)
			(page-menu (xmltrans:get-parent "PAGE")))))
	 (xmltrans:expect-next-tag "HEADER" proc "NOTREADY"))))
 (xmltrans:set-attr #f "MENU" '())
 #f)
	
(xmltrans:end-tag
 "PAGE"
 (tag attr text)
 (gamlet:with-output-to-file
  (current-file-name)
  (lambda ()
    (open-page (page-title))
    (display text)
    (close-page)))
 #f)

;;; KEYWORDS
(xmltrans:end-tag
 "KEYWORDS"
 (tag attr text)
 (set! keywords text)
 "")

;;; TITLE
(xmltrans:end-tag
 "TITLE"
 (tag attr text)
 (cond
  ((xmltrans:parent? "TABULAR")
   (string-append "<caption>" text "</caption>"))
  (else
   (xmltrans:parse-error #f "Misplaced TITLE")
   #f)))

;;; CHAPTER
(xmltrans:start-tag
 "CHAPTER"
 (tag attr)
 (xmltrans:set-attr #f "PAGENO" 0)
 (xmltrans:set-attr #f "MENU" '())
 #f) ;; Cannot have (set-anchor attr))

(xmltrans:end-tag
 "CHAPTER"
 (tag attr text)  
 (menu-append #t (xmltrans:attr attr "MENU") "")
 #f)

;;; GROUP
(xmltrans:start-tag
 "GROUP"
 (tag attr)
 (xmltrans:set-attr #f "MENU" '())
 #f)

(xmltrans:end-tag
 "GROUP"
 (tag attr text)
 (if (not (assoc "HIDDEN" attr))
     (menu-append #t
		  (xmltrans:attr attr "MENU")
		  (xmltrans:attr attr "TITLE")))
 #f)



(define (check-sectioning parent)
  (if (not (xmltrans:parent? parent))
      (xmltrans:parse-error #f "Incorrect sectioning. Expected "
			    parent
			    " but found "
			    (xmltrans:get-parent))))

;;; SECTION
(xmltrans:start-tag
 "SECTION"
 (tag attr)
 (check-sectioning "PAGE")
 (set-anchor attr))

(xmltrans:end-tag
 "SECTION"
 (tag attr text)
 (list "<div class=\"section\">" text "</div>"))

;;; SUBSECTION
(xmltrans:start-tag
 "SUBSECTION"
 (tag attr)
 (check-sectioning "SECTION")
 (set-anchor attr))

(xmltrans:end-tag
 "SUBSECTION"
 (tag attr text)
 (list "<div class=\"subsection\">" text "</div>"))

;;; SUBSUBSECTION
(xmltrans:start-tag
 "SUBSUBSECTION"
 (tag attr)
 (check-sectioning "SUBSECTION")
 (set-anchor attr))

(xmltrans:end-tag
 "SUBSUBSECTION"
 (tag attr text)
 (list "<div class=\"subsubsection\">" text "</div>"))

;;; ANCHOR
(xmltrans:end-tag
 "ANCHOR"
 (tag attr text)
 (set-anchor attr text))

;;; XREF
(xmltrans:start-tag
 "XREF"
 (tag attr)
 (cond
  ((assoc-ref attr "THIS") =>
   (lambda (expr)
     (format #f "<a href=\"~A\" class=\"this\">"
	     (eval
	      (with-input-from-string expr
	      read)
	      (resolve-module '(xmltrans lingua))))))
  ((assoc-ref attr "LIT") =>
   (lambda (text)
     (format #f "<a href=\"~A\" class=\"lit\">" text)))
  ((assoc "URL" attr) =>
   (lambda (url)
     (format #f "<a href=\"~A\" class=\"external\">" (cdr url))))
  ((assoc "REF" attr) =>
   (lambda (ref)
     (let* ((ref-name (cdr ref))
	    (id (xref-retrieve ref-name)))
       (cond
	((and id (xref-defined? id))
	 (string-append
	  "<a href=\""
	  (reference-name (file-name->reference (xref-file-name id)))
	  "#"
	  ref-name
	  "\">"))
	(else
	 (if (not id) 
	     (xref-register ref-name #f))
	 ;; The reference is registered but not defined
	 ;; Create a postprocess handler and insert its invocation to the
	 ;; output stream. The handler is defined within a closure that
	 ;; supplies target reference name and source location for
	 ;; resolve-forward-xref
	 (let ((location (xmltrans:get-input-location)))
	   (letrec ((name (string-append "@@" ref-name ":"))
		    (proc (lambda ()
			    (resolve-forward-xref ref-name location))))
	     (gamlet:set-handler name proc)
	     (let ((delim (gamlet:get-pp-delimiter)))
	       (format #f "<a href=\"~A~A~A\">"
		       (car delim) name (cdr delim))))))))))
  ((assoc "SOUND" attr) =>
   (lambda (val)
     (if (and
	  (not (lingua:tag-exists? "SOUNDREF"))
	  (memq #:soundref warning-list))
	 (xmltrans:parse-warning #f
			       "<XREF SOUND> used but <SOUNDREF> is not defined"))
     (string-append
      "<a href=\""
      (or (and (lingua:tag-exists? "DIRECTORY") (lingua:attr "DIRECTORY" "SOUND")) "")
      "/"
      (cdr val)
      "\" class=\"sound\">")))
  (else
   (xmltrans:parse-error #f "XREF element must have URL, REF or SOUND attrbute")
   "<a href=\"NONEXISTENT\">")))

(xmltrans:end-tag
 "XREF"
 (tag attr text)
 (list text
       (or (and (assoc "SOUND" attr)
		(lingua:tag-exists? "SOUNDREF")
		(lingua:tag-text "SOUNDREF")) "")
       "</a>"))

;;; BOTTOM	     
(xmltrans:end-tag
 "BOTTOM"
 (tag attr text)
 (xmltrans:set-attr "PAGE" "BOTTOM" text)
 #f)

(define (footnote-parent)
  (if (xmltrans:tag-exists? "TABULAR") ;; FIXME
      "TABULAR"
      "PAGE"))

;;; FOOTNOTE
(xmltrans:end-tag
 "FOOTNOTE"
 (tag attr text)
 (cond
  ((not (xmltrans:tag-exists? "PAGE"))
   (xmltrans:parse-error #f
			 "Incorrect sectioning. FOOTNOTE may be used only inside PAGE"))
  (else
   (let ((n (add-footnote (footnote-parent) text (xmltrans:attr attr "ID"))))
     (format #f "<a name=\"~A\" href=\"~A#~A\"><sup>~A)</sup></a>"
	     (footnote-name n)
	     (reference-name (file-name->reference (current-file-name)))
	     (footnote-href n)
	     n)))))

;;; PARA
(xmltrans:start-tag
 "PARA"
 (tag attr)
 "\n<p>")

(xmltrans:end-tag
 "PARA"
 (tag attr text)
;;FIXME: This may generate empty <P> blocks
 (if (not (assoc "__CLOSED__" attr))
     (list text "</p>\n")
     text))

;;; HEADER

(define (format-header level text)
  (let ((pfx (set-anchor #f)))
    (format #f "\n~A<h~A>~A</h~A>\n"
	    (or pfx "") level text level)))

(xmltrans:end-tag
 "HEADER"
 (tag attr text)
 (cond
  ((xmltrans:parent? "PAGE")
   (cond
    ((xmltrans:attr #t "HEADER")
     (if (and (not (xmltrans:attr #t "__HEADER_INSERTED__"))
	      (memq #:page-header warning-list))
	 (xmltrans:parse-warning #f
				 "PAGE element already has HEADER defined"))
     (xmltrans:set-attr "PAGE" "HEADER" text))
    (else
     (xmltrans:set-attr "PAGE" "HEADER" text)
     (page-menu (xmltrans:get-parent "PAGE"))))
    (set-anchor attr))
  ((xmltrans:parent? "SECTION")
   (format-header 2 text))
  ((xmltrans:parent? "SUBSECTION")
   (format-header 3 text))
  ((xmltrans:parent? "SUBSUBSECTION")
   (format-header 4 text))
  ((xmltrans:parent? "DEFLIST")
   (list "<dt>" text "</dt>"))
  (else
   (xmltrans:parse-error #f "Misplaced HEADER")
   #f)))

;;; 
;;; Split the given text string at split delimiters. Print each part
;;; except the first one in bold face and indent it one space from
;;; the previous one  
(define (item-split text)
  (cond
   ((xmltrans:attr "TABULAR" "SPLIT") =>
    (lambda (delim)
      (let ((l (string-split text (string-ref delim 0))))
	(if (null? (cdr l))
	    (car l)
	    (apply string-append
		   (cons
		    (car l)
		    (do ((cur (cdr l) (cdr cur))
			 (out '()))
			((null? cur)
			 out)
		      (set! out (append out (list "&nbsp;<b>" (car cur) "</b>"))))))))))
   (else
    text)))

;;; Print the contents of the current item using splitting rules.
;;;  SPLIT attribute introduces a split delimiter.
;;;  OR attribute introduces an "or" delimiter
(define (item-text text)
  (cond
   ((xmltrans:attr "TABULAR" "OR") =>
    (lambda (delim)
      (let ((l (string-split text (string-ref delim 0))))
	(if (null? (cdr l))
	    (item-split text)
	    (apply string-append
		   (cons
		    (item-split (car l))
		    (do ((cur (cdr l) (cdr cur))
			 (out '()))
			((null? cur)
			 out)
		      (set! out
			    (append
			     (list "&nbsp;" delim "&nbsp;"
				   (item-split (car cur))))))))))))
   (else
    (item-split text))))

;;; Remove any split delimiters from the given text string	   
(define (restore-single-word text)
  (cond
   ((xmltrans:attr "TABULAR" "SPLIT") =>
    (lambda (delim)
      (apply string-append (string-split text (string-ref delim 0)))))
   (else
    text)))

;;; Restore the "canonical" form of the item text by removing any split
;;; delimiters from it and gluing the resulting parts together. If the
;;; item has several parts separated by "or" delimiters, only the first
;;; part is acted upon.
(define (restore-word text)
  (cond
   ((xmltrans:attr "TABULAR" "OR") =>
    (lambda (delim)
      (restore-single-word (car (string-split text (string-ref delim 0))))))
   (else
    (restore-single-word text))))

;;; Create a dictionary reference for a given WORD. REFTEXT gives the
;;; reference text.
(define (dict-reference word reftext . rest)
  (let ((backref (and (not (null? rest)) (car rest))))
    (format #f
	    "<a ~Ahref=\"~A&amp;ident=~A\">~A</a>"
	    (if backref
		(string-append "name=\"id" (dict:encode-string word) "\" ")
		"")
	    ((xmltrans:attr "LINGUA" "DICT"))
	    (dict:encode-string word)
	    reftext)))

;;; Return true if we are on the very first data row in a table. Colon
;;; headings are not counted as data row.

(define (first-data-row?)
  (if (xmltrans:attr "TABULAR" "COLHEADING")
      (= row-index 1)
      (= row-index 0)))

;;; Return effective row index
;;; Rows are numbered from 1, column headings are not counted	   
(define (eff-row-index)
  (if (xmltrans:attr "TABULAR" "COLHEADING")
      row-index
      (1+ row-index)))

;;; Return effective column (item) index
;;; Columns are numbered from 1, row headings are not counted	   
(define (eff-col-index)
  (if (xmltrans:attr "TABULAR" "ROWHEADING")
      item-index
      (1+ item-index)))

;;; Table map structure: #( start-col end-col start-row end-row )
;;; FIXME: Originally it had 4th element: need-ref?
(define-macro (map-start-col map)
  `(vector-ref ,map 0))

(define-macro (map-end-col map)
  `(vector-ref ,map 1))

(define-macro (map-start-row map)
  `(vector-ref ,map 2))

(define-macro (map-end-row map)
  `(vector-ref ,map 3))

(define (parse-dictref str)
  (let ((map (vector -1 -1 -1 -1)))
    (letrec ((set-map (lambda (n x)
			(vector-set! map n (string->number (car x)))
			(vector-set! map (1+ n)
				     (if (or (null? (cdr x))
						    (string-null? (cadr x)))
					 0
					 (string->number (cadr x)))))))
      (for-each
       (lambda (s)
	 (let ((x (string-split s #\:)))
	   (cond
	    ((string=? (car x) "col")
	     (set-map 0 (cdr x)))
	    ((string=? (car x) "row")
	     (set-map 2 (cdr x)))
	    (else
	     (if (memq #:unknown-modifier warning-list)
		 (xmltrans:parse-warning #f "Unknown modifier " x))))))
       (string-split str #\;))
      (cond
       ((and (= (vector-ref map 0) -1) (not (= (vector-ref map 2) -1)))
	(vector-set! map 0 1)
	(vector-set! map 1 1))
       ((and (= (vector-ref map 2) -1) (not (= (vector-ref map 0) -1)))
	(vector-set! map 2 1)
	(vector-set! map 3 1)))
      map)))

(define (need-dictref)
  (cond
   ((xmltrans:attr "TABULAR" "DICTREF") =>
    (lambda (x)
      (let ((map (parse-dictref x)))
	(and
	 (not (= (map-start-col map) -1))
	 (or
	  (= (map-start-col map) 0)
	  (<= (map-start-col map) (eff-col-index)))
	 (or
	  (= (map-end-col map) 0)
	  (>= (map-end-col map) (eff-col-index)))
	 (not (= (map-start-row map) -1))
	 (or
	  (= (map-start-col map) 0)
	  (<= (map-start-row map) (eff-row-index)))
	 (or
	  (= (map-end-row map) 0)
	  (>= (map-end-row map) (eff-row-index)))))))
   (else
    #f)))

(define (get-col-defs ind attr-name)
  (let ((cl (xmltrans:attr "TABULAR" attr-name)))
    (if cl
	(let ((s (string-split cl #\:)))
	  (cond
	   ((null? (cdr s))
	    cl)
	   ((<= (length s) ind)
	    (if (string=? (car (last-pair s)) "*")
		(list-ref s (- (length s) 2))
		#f))
	   ((string=? (list-ref s ind) "*")
	    (list-ref s (- ind 1)))
	   ((string-null? (list-ref s ind))
	    #f)
	   (else
	    (list-ref s ind))))
	#f)))

(xmltrans:end-tag
 "ITEM"
 (tag attr text)
 (cond
  ((or (xmltrans:parent? "ENUMERATE") (xmltrans:parent? "ITEMIZE"))
   (list "<li>" text "</li>"))
  ((xmltrans:parent? "DEFLIST")
   (list "<dd>" text "</dd>"))
  ((xmltrans:parent? "ROW")
   (letrec ((sl (list))
	    (put (defn-put put sl)))
     (let ((eltname (if (and
			 (= row-index 0)
			 (xmltrans:attr "TABULAR" "COLHEADING"))
			"th"
			"td")))

       (put "<" eltname)
       (let ((span (xmltrans:attr "ITEM" "SPAN")))
	 (cond
	  (span
	   (let ((n (if (string=? span "0")
			max-items
			span)))
	     (put " colspan=\"" span "\" class=\"center\"")))
	  (else
	   (cond
	    ((string=? eltname "td")
	     (let ((align (get-col-defs item-index "COLALIGN")))
	       (if align
		   (put " align=\"" (string-downcase align) "\"")))

	     (let ((class (get-col-defs item-index "COLCLASS")))
	       (if class
		   (class " class=\"" (string-downcase class) "\"")))
	     
	     (let ((rowspan (xmltrans:attr "ITEM" "ROWSPAN")))
	       (cond
		((and (= item-index 0) rowspan)
		 (set! rowspan-end (+ row-index
				      (string->number rowspan)))
		 (put " rowspan=\"" rowspan "\""))))))))

	 (let ((rowheading (and (= item-index 0)
				(xmltrans:attr "TABULAR" "ROWHEADING"))))
	   (if rowheading
	       (put " class=\"" rowheading "\""))

	   (put ">")

	   (cond
	    (rowheading
	     (put text))
	    (else
	     (let ((x (need-dictref)))
	       (if x
		   (put (dict-reference (restore-word text)
					(item-text text)
					x))
		   (put (item-text text))))))
	   (put "</" eltname ">")
	   (set! item-index (1+ item-index))
	   (if (< max-items item-index)
	       (set! max-items item-index)))))

     (reverse sl)))
  ((xmltrans:parent? "GROUP")
   (let ((fn (xmltrans:attr "LINGUA" text)))
     (cond
      ((not fn)
       (xmltrans:parse-error #f "Function " text " not defined"))
      ((not (procedure? fn))
       (xmltrans:parse-error #f text " is not a function"))
      (else
       (menu-append #t fn ""))))
   #f)
  ((xmltrans:parent? "PAGE")
   (list "<h2>" text "</h2>"))
  (else
   (list "<dd>" text "</dd>"))))

(xmltrans:end-tag
 "FLECT"
 (tag attr text)
 (cond
  ((xmltrans:parent? "HEADER")
   text)
  ((assoc-ref attr "SPLIT") =>
   (lambda (delim)
     (let ((sl (list)))
       (letrec ((put (defn-put put sl)))
	 (put "<b>")
	 (let ((tok-list (string-split text (string-ref delim 0))))
	   (put "-" (car tok-list))
	   (for-each
	    (lambda (s)
	      (put "&nbsp;,&nbsp;-" s))
	    (cdr tok-list)))
	 (put "</b>")
	 (reverse sl)))))
  (else
   (list "<b>-" text "</b>"))))

(xmltrans:end-tag
 "ALPH"
 (tag attr text)
 (list "<span class=\"literal\">" text "</span>"))

(xmltrans:end-tag
 "IPA"
 (tag attr text)
 (let ((ret (list "<span class=\"ipa\">" text "</span>"))
       (sound (assoc "SOUND" attr)))
   (if sound
       (append
	(list
	 "<a href=\""
	 (or (and (lingua:tag-exists? "DIRECTORY") (lingua:attr "DIRECTORY" "SOUND")) "")
	 "/"
	 (cdr sound)
	 "\" class=\"sound\">"
	ret
	(list
	 (or (and (lingua:tag-exists? "SOUNDREF") (lingua:tag-text "SOUNDREF")) ""))
	 "</a>"))
       ret)))

(xmltrans:end-tag
 "DICTREF"
 (tag attr text)
 (dict-reference text text))

;;; ENUMERATE
(xmltrans:end-tag
 "ENUMERATE"
 (tag attr text)
 (list
  (cond
   ((xmltrans:parent? "PARA")
    (xmltrans:set-attr "PARA" "__CLOSED__" 1)
    "\n</p>\n")
   (else
    ""))
  "<ol>\n" text "</ol>\n"))

(xmltrans:end-tag
 "ITEMIZE"
 (tag attr text)
 (list
  (cond
   ((xmltrans:parent? "PARA")
    (xmltrans:set-attr "PARA" "__CLOSED__" 1)
    "\n</p>\n")
   (else
    ""))
  "<ul>\n" text "</ul>\n"))


(xmltrans:end-tag
 "DEFLIST"
 (tag attr text)
 (list
  (cond
   ((xmltrans:parent? "PARA")
    (xmltrans:set-attr "PARA" "__CLOSED__" 1)
    "\n</p>\n")
   (else
    ""))
  "<dl>\n" text "</dl>\n"))

(xmltrans:start-tag
 "TABULAR"
 (tag attr)
 (letrec ((sl (list))
	  (put (defn-put put sl)))
   (if (xmltrans:parent? "PARA")
       (begin
	 (xmltrans:set-attr "PARA" "__CLOSED__" 1)
	 (put "\n</p>\n")))
   
   (let ((sp (xmltrans:attr #f "SPACE")))
     (if sp
	 (do ((i 0 (1+ i)))
	     ((= i (string->number sp)))
	   (put "<br />"))))

   (if (and (memq #:colalign-used warning-list)
	    (xmltrans:attr #f "COLALIGN"))
       (xmltrans:parse-warning #f "The use of COLALIGN is deprecated"))
   
   (put "<table")

   (let ((type (or (xmltrans:attr #f "TYPE") ""))
	 (class #f))
     (let ((add-class (lambda (cl)
			(set! class
			      (if (not class)
				  cl
				  (string-append class " " cl))))))
			
       (cond
	((string=? type "EXAMPLE")
	 (xmltrans:set-attr #f "NOFRAME" 1)))

       (if (xmltrans:attr #f "NOFRAME")
	   (set! class "noframe")
	   (set! class "frame"))
;     (if (not (xmltrans:attr #f "BESTFIT"))
;	 (put " width=\"95%\""))

       (cond
	((xmltrans:attr #f "CLASS") => add-class))

       (cond
	((xmltrans:attr #f "ALIGN") =>
	 (lambda (align)
	   (if (not (string-ci=? align "NONE"))
	       (add-class (string-append "align-" (string-downcase align))))))
	(else
	 (add-class "align-center")))

       (if  (not (null? class))
	    (put " class=\"" class "\""))

       (put ">\n")

       (if (string=? type "EXAMPLE")
	   (put "<col span=\"2\" width=\"50%\" />\n"))

       (set! row-index 0)
       (set! max-items 0)
       (set! separator-row 0)

       (reverse sl)))))

(define table-number 0) ;; Ordinal number of the current table

(xmltrans:end-tag
 "TABULAR"
 (tag attr text)
 (set! table-number (1+ table-number))
 (let ((sl (list)))
   (letrec ((put (defn-put put sl)))
     (put text)
     (put "</table>\n")

     (let ((sp (xmltrans:attr #f "AFTERSPACE")))
       (if sp
	   (do ((i 0 (1+ i)))
	       ((= i (string->number sp)))
	     (put "<br />"))))

     (let ((x (xmltrans:attr #f "__COLSPAN__")))
       (if x
	   (let ((max max-items))
	     (letrec ((proc (lambda ()
			      (display max))))
	       (gamlet:set-handler x proc)))))

     (let ((fl (xmltrans:attr #f "FOOTNOTES")))
       (map-footnotes fl put))

     (reverse sl))))

;;; ROW

(xmltrans:start-tag
 "ROW"
 (tag attr)
 (let ((sl (list)))
   (letrec ((put (defn-put put sl)))
     (set! item-index 0)
     (cond
      ((and (= row-index 0) (xmltrans:attr "TABULAR" "COLHEADING")) =>
       (lambda (class)
	 (put "<thead class=\"" class "\">" "<tr>\n")))
      (else
       (put "<tr")
       (if (xmltrans:attr "TABULAR" "ALTERNATE")
	   (put " class=\"" (if (= (modulo (- row-index separator-row) 2) 0)
				"even"
				"odd")
		"\""))
       (put ">\n")))
     (if (< row-index rowspan-end)
	 (set! item-index (1+ item-index)))
     (reverse sl))))

(xmltrans:end-tag
 "ROW"
 (tag attr text)
 (let ((retval (list text "</tr>")))
   (if (and (= row-index 0)
	    (xmltrans:attr "TABULAR" "COLHEADING"))
       (set! retval (append retval (list "</thead>"))))

   (set! row-index (1+ row-index))
   (if (= row-index rowspan-end)
       (set! rowspan-end 0))
   
   (append retval (list "\n"))))
   

(define separator-count 0) ;; Number of separators for which a GAMLET macro
                           ;; has been created.

;;; SEPARATOR   
(xmltrans:start-tag
 "SEPARATOR"
 (tag attr)
 (let ((max (cond
	     ((xmltrans:attr #f "SPAN") =>
	      (lambda (x)
		x))
	     ((= row-index 0)
	      ;; This is the first row. We still do not know how many columns
	      ;; will it have.
	      (let ((x (xmltrans:attr "TABULAR" "__COLSPAN__")))
		(if x
		    x
		    (let ((name (string-append
				 "@@COLSPAN:"
				 (number->string separator-count))))
		      (set! separator-count (1+ separator-count))
		      (xmltrans:set-attr "TABULAR" "__COLSPAN__" name)
		      (let ((delim (gamlet:get-pp-delimiter)))
			(format #f "~A~A~A" (car delim) name (cdr delim)))))))
	     (else
	      (number->string max-items)))))
   (list "\n<tr><td colspan=\"" max "\" class=\"separator\">")))

(xmltrans:end-tag
 "SEPARATOR"
 (tag attr text)
 (set! separator-row (if (or (xmltrans:attr #f "INITCOUNT")
			     (xmltrans:attr "TABULAR" "INITCOUNT"))
			 0
			 row-index))
 (list text "</td></tr>\n"))
     

;;;

(define (class-begin attr class-name)
  (letrec ((sl (list))
	   (put (defn-put put sl)))
    (let ((tag (cond
		((xmltrans:parent? "ROW")
		 "td")
		((xmltrans:parent? "TABULAR")
		 (put "<tr class=\"" class-name "\""))
		((xmltrans:attr attr "TYPE") =>
		 (lambda (type)
		   (if (string=? type "INLINE")
		       "span"
		       "div")))
		((xmltrans:attr attr "ALIGN") =>
		 (lambda (x)
		   (string-append "div align=\"" x "\"")))
		(else
		 "span"))))
      (put "<" tag " class=\"" class-name "\">")
      (reverse sl))))

(define (class-end attr)
  (letrec ((sl (list))
	   (put (defn-put put sl)))
    (let ((tag (cond
		((xmltrans:parent? "ROW")
		 "td")
		((xmltrans:parent? "TABULAR")
		 "td")
		((xmltrans:attr attr "TYPE") =>
		 (lambda (type)
		   (if (string=? type "INLINE")
		       "span"
		       "div")))
		((xmltrans:attr attr "ALIGN")
		 "div")
		(else
		 "span"))))
      (put "</" tag ">")
      (if (xmltrans:parent? "TABULAR")
	  (put "</tr>"))
      (reverse sl))))

(xmltrans:end-tag
 "EXAMPLE"
 (tag attr text)
 (let ((type (xmltrans:attr attr "TYPE")))
   (if (and type (not (string=? type "INLINE")))
       (list
	"<div class=\"center\">"
	(class-begin #f "example")
	text
	(class-end #f)
	"</div>")
       ; else
       (list
	(class-begin #f "example")
	text
	(class-end #f)))))

(xmltrans:end-tag
 "TRANS"
 (tag attr text)
 (list
  (class-begin #f "translation")
  text
  (class-end #f)))

(xmltrans:end-tag
 "SAMP"
 (tag attr text)
 (list
  "<span class=\"sample\">"
  text
  "</span>"))

(xmltrans:end-tag
 "NOTREADY"
 (tag attr text)
 (let ((ref (image-reference)))
   (list
    (format #f "
 <table class=\"notready\">
  <tr>
   <td><img src=\"~A/construction.jpg\" alt=\"[ Construction Bar ]\" width=\"300\" height=\"28\" /></td>
<td class=\"notready\">"
	    ref)
     text
    (format #f "
   </td>
   <td><img src=\"~A/construction.jpg\" alt=\"[ Construction Bar ]\" width=\"300\" height=\"28\" /></td>
   </tr>
</table>"
	    ref))))


;;;
(define (alignment a)
  (if (string-null? a)
      "</div>\n"
      (string-append "<div class=\"" (string-downcase a) "\">")))


;;; Postprocessor handlers

(define (toc-pp)
  (out-menu (lingua:attr "LINGUA" "MENU") 0))

(gamlet:set-handler "TOC" toc-pp)

(define (sitemap-pp)
  (print-menu-map (lingua:attr "LINGUA" "MENU")))

(gamlet:set-handler "SITEMAP" sitemap-pp)

(gamlet:set-handler "INFO" 
;; Empty function. The user is supposed to override it if he needs to
         (lambda (. args)
           #f))
  

(define (bottom-navbar-pp)
  (display "<div class=\"bottom-navbar\">")
  (let ((index (gamlet:output-file-index)))
    (display "<table class=\"bottom-navbar\" width=\"100%\">
<tr>
 <td class=\"prev-link\">")
    (if (> index 0)
	(format #t "<a href=\"~A\"><img src=\"~A/left.png\" alt=\"[ ~A ]\" />~A</a>"
		(reference-name
		 (file-name->reference (gamlet:output-file-name (- index 1))))
		(image-reference)
		(lingua:gettext "Previous Page")
		(find-file-in-menu
		 (file-name->reference (gamlet:output-file-name (- index 1))))))

    (display "</td><td class=\"top-link\">")
    (format #t "<a href=\"~A#top\"><img src=\"~A/up.png\" alt=\"[ ~A ]\" /><br />~A</a>"
	    (reference-name
	     (file-name->reference (gamlet:current-output-file-name)))
	    (image-reference)
	    (lingua:gettext "Top of the Page")
	    (lingua:gettext "Top of the Page"))

    (display "</td><td class=\"next-link\">")
    
    (if (< index (- (gamlet:output-file-count) 1))
	(format #t "<a href=\"~A\">~A<img src=\"~A/right.png\" alt=\"[ ~A ]\" /></a>"
		(reference-name
		 (file-name->reference (gamlet:output-file-name (+ index 1))))
		(find-file-in-menu
		 (file-name->reference (gamlet:output-file-name (+ index 1))))
		(image-reference)
		(lingua:gettext "Next Page")))

    (display "</td>
 </tr> 
</table>
</div>")))

(gamlet:set-handler "BOTTOM-NAVBAR" bottom-navbar-pp)      


;;;; Command line parser
(define (cons? p)
  (and (pair? p) (not (list? p))))

(define (usage)
  (display "usage: xmlt lingua OPTIONS\n")
  (display "OPTIONS are:\n")
  (display "  -b, --base-href URL\n")
  (display "  -p, --prefix DIR\n")
  (display "  -c, --config-file FILE\n")
  (display "  -o, --output-dir DIR\n")
  (display "  -t, --tmp-dir DIR\n")
  (display "  -d, --debug NUMBER\n")
  (display "  -D, --define SYM[=VAL]\n")
  (display "  -U, --undefine SYM\n")
  (display "  -W, --warning [no-]WCLASS (WCLASS is one of prereq, author-data, page-title, page-header, unknown-modifier, colalign-used, soundref\n")
  (display "  -V, --version\n")
  0)

(define (version)
  (format #t "~A~%" lingua-version-string)
  (display lingua-copyright)
  (exit 0))

(define (parse-command-line arglist)
  (let ((grammar `((base-href (single-char #\b)
			      (value #t))
		   (config-file (single-char #\c)
				(value #t))
		   (prefix    (single-char #\p)
			      (value #t))
		   (output-dir (single-char #\o)
			       (value #t))
		   (tmp-dir   (single-char #\t)
			      (value #t))
		   (debug (single-char #\d)
			  (value #t))
		   (warning (single-char #\W)
			    (value #t))
		   (define (single-char #\D)
		           (value #t))
		   (undefine (single-char #\U)
			     (value #t))
		   (help (single-char #\h))
		   (version (single-char #\V))))
	(input-files '()))
    (for-each
     (lambda (x)
       (cond
	((cons? x)
	 (case (car x)
	   ((debug)
	    (xmltrans:set-debug-level (string->number (cdr x))))
	   ((config-file)
	    (set! config-must-exist #t)
	    (set! config-file-name (cdr x)))
	   ((help)
	    (exit (usage)))
	   ((version)
	    (version))
	   ((prefix)
	    (set! prefix-option (cdr x)))
	   ((base-href)
	    (set! base-href-option (cdr x)))
	   ((output-dir)
	    (set! output-dir-option (cdr x)))
	   ((tmp-dir)
	    (set! tmp-dir-option (cdr x)))
	   ((define)
	    (let ((vp (let* ((rval (cdr x))
			     (n (string-index rval #\=)))
			(if n
			    (cons (substring rval 0 n)
				  (substring rval (1+ n)))
			    (cons rval #t)))))
	      (lingua:define (car vp) (cdr vp))))
	   ((undefine)
	    (lingua:undefine (cdr x)))
	   ((warning)
	    (set-warning-control (cdr x)))))
	(else
	 (set! input-files (cdr x)))))
     (getopt-long arglist grammar))
    input-files))

(define (begin-processing)
  ;; First, see if we have a configuration file
  (cond
   ((file-exists? config-file-name)
    (if (not (xmltrans:parse-file config-file-name))
	(exit 1)))
   (config-must-exist
    (display (string-append
	      "Configuration file " config-file-name " not found\n")
	     (current-error-port))
    (exit 1)))

  (cond
   ((or output-dir-option
	(and (lingua:tag-exists? "DIRECTORY")
	     (lingua:attr "DIRECTORY" "OUT"))) =>
	     (lambda (x)
	       (gamlet:set-output-dir x))))

  (cond
   ((or tmp-dir-option
	(and (lingua:tag-exists? "DIRECTORY")
	     (lingua:attr "DIRECTORY" "TMP"))) =>
	     (lambda (x)
	       (gamlet:set-working-dir x)))))

(define (end-processing)
  #t)

;;;; End of lingua.scm

   
